/**
 * Created by senthil on 2/2/2017.
 */
var elasticsearch = require('elasticsearch');
var _ = require('lodash');
var async = require('async');
var base64Img = require('base64-img');
var uuid = require('uuid');
var esConstants = require('./constants');
var guidelineJson = require('./guideline-config.json');

var defaultOptions = {
    hosts: [],
    disabled_hosts: [],
    depreciatedLogging: (process.env.NODE_ENV === undefined || 'development') ? console.log : false,
    protocol: 'http',
    host: 'localhost',
    port: 9200,
    username: 'admin',
    password: 'admin',
    index: 'perfx360accessibility',
    indexType: 'ACCESSIBILITY',
    summaryIndexType: 'ACCESSIBILITY_SUMMARY',
    failoverTimeout: 60000,
    requestTimeout: null,
    maxRetries: 2,
    timePrecision: 'ms',
    index: ''
}


var principles = {
    "Principle1" : "Perceivable",
    "Principle2" : "Operable",
    "Principle3" : "Understandable",
    "Principle4" : "Robust"
}

var sc = {
    "WCAG2A" : {"sc" : 25, "total": 25},
    "WCAG2AA" : {"sc" : 13, "total": 38},
    "WCAG2AAA" : {"sc" : 23, "total": 61}
}

var SECTION508 = "Section508";

var client = function f(options) {
    var self = this;
};

var ElasticClient = function (options) {
    this.options = _.extend(_.clone(defaultOptions), options)

    this.es = new elasticsearch.Client({
        host: options.protocol + "://" + options.host + ':' + options.port
    });

    return this;
}

ElasticClient.prototype.getClient = function () {
    return this.client;
}

ElasticClient.prototype.indexExists = function(callback) {
    isIndexExists(this.es, { index: this.options.index }, function(err, response) {
        callback(err, response);
    })
}

ElasticClient.prototype.ping = function(callback) {
    ping(this.es, {requestTimeout: Infinity}, function(err, response, status) {
        callback(err, response, status);
    })
}

ElasticClient.prototype.putMapping = function (cb) {
    var params = {};
    params.options = this.options;
    params.es = this.es;

    async.waterfall([
        async.apply(_createMapping, params),
        _createSummaryMapping
    ], function (err, params) {
        cb(err, params);
    });
}

ElasticClient.prototype.publishDocument = function (params, cb) {
    params.es = this.es;
    params.options = this.options
    params.index = this.options.index;
    params.indexType = this.options.indexType;

        async.waterfall([
        async.apply(_getDocument, params),
        _searchSummaryDocument,
        _getSummaryDocument,
        _publishDocument
    ], function (err, params) {
        cb(err, params);
    });
}

ElasticClient.prototype.getA11ySummaryResults = function (params, cb) {
    params.es = this.es;
    params.options = this.options
    params.index = this.options.index;
    params.indexType = this.options.summaryIndexType;

    async.waterfall([
        async.apply(getAllySummaryResult, params),
    ], function (err, params) {
        cb(err, params);
    });
}

ElasticClient.prototype.createSummaryDocument = function (params, cb) {
    params.es = this.es;
    params.options = this.options
    params.index = this.options.index;
    params.indexType = this.options.indexType;

    async.waterfall([
        async.apply(_createSummaryDocument, params),
        _publishDocument
    ], function (err, params) {
        cb(err, params);
    });
}

function _createMapping(params, callback) {
    var mappingParam = getMapping(params.options);
    params.es.indices.create(mappingParam, function(error, response) {
        params.response = response;
        callback(error, params);
    });
}

function _createSummaryMapping(params, callback) {
    var summaryMappingParam = getSummaryMapping(params.options);
    params.es.indices.create(summaryMappingParam, function(err, response) {
        params.response = response;
        callback(err, params);
    });
}

function _searchSummaryDocument(params, callback) {
    var searchParam = getSummarySearchParam(params);
    params.es.search(searchParam, function(err, response) {
        params.summaryDocument = response;
        callback(err, params);
    });
}


function _getDocument(params, callback) {
    var points = [],
        samples = 0,
        sample;
    var constants = params.constants;
    var instance = params.instance;
    var testReport = params.testReport;
    var results = testReport.result;
    var documents = [];

    // For each of the latencies, create document for elasticsearch,
    // adding any static tags from the config.

    //var hitId = Math.floor((Math.random() * 10000000000000) + 1);

    console.info('params... = ', params);
    if (results.length > 0) {
        params.sectionId = uuid.v4();

        for (var i = 0; i < results.length; i++) {
            var result = results[i];
            var document = {};
            document.testExecutionId = instance.config[constants.CONFIG_TEST_EXECUTION_ID];
            document.userId = instance.config[constants.CONFIG_TEST_USER_ID];
            document.userName = instance.config[constants.CONFIG_TEST_USER_NAME];
            document.projectId = instance.config.projectId;
            document.componentId = instance.config.applicationId;
            document.projectName = instance.config.projectName;
            document.projectDescription = instance.config.projectDescription;
            document.testId = instance.config.testId;
            document.testName = instance.config.testName;
            document.environmentId = instance.config.environmentId;
            document.environmentName = instance.config.environmentName;
            document.testExecutionName = instance.config.testExecutionName;
            document.measurementName = instance.config.measurementName;
            if (result.code) {

                var codeRes = result.code;
                document.code = codeRes.replace(/\./g, '-');
                var docDetails = _.split(result.code, '.', 8);

                if(docDetails[1]) {
                    document.principle = principles[docDetails[1]];
                    document.principleSubTitle = getPrincipleSubTitle(docDetails[3]);
                }

                if(docDetails[2]) {
                    document.guideline = docDetails[2];
                }

                if(docDetails[3]) {
                    if ('1_4_3_F24' == docDetails[3]) {
                        document.guidelineIndex = "1_4_3";
                    } else if ('1_3_1_A' == docDetails[3]) {
                        document.guidelineIndex = "1_3_1";
                    }  else {
                        document.guidelineIndex = docDetails[3];
                    }
                }

                var guideline = _.find(guidelineJson, {guidelineIndex: document.guidelineIndex});
                //console.info('******** guideline = ', guideline);

                if(guideline) {
                    document.title = guideline.title;
                    document.description = guideline.description;
                    document.testing = guideline.testing;
                    document.level = guideline.level;
                }

                if(!document.level) {
                    console.info('******** if document.level undefined = ', docDetails[3]);
                } else {
                    console.info('******** else document.level = ', guideline);
                }

                if(docDetails[4]) {
                    var techniques = docDetails[4].split(",");
                    var temp = [];
                    if (techniques) {
                        for (var j = 0; j < techniques.length; j++) {
                            temp[j] = techniques[j].toString();
                        }
                    }
                    document.techniques = temp;
                    document.techniqueName = getTechnique(docDetails[4]);
                }

                if(docDetails[5]) {
                    document.object = docDetails[5];
                }
            }

            if (testReport.requestUrl) {
                document.test_url = testReport.requestUrl.url;
            }

            if (testReport.standard) {
                document.standard = testReport.standard;
            }

            if (result.context) {
                document.context = result.context;
            }

            if (result.message) {
                document.message = result.message;
            }

            if (result.selector) {
                document.selector = result.selector;
            }

            if (result.type) {
                document.type = result.type;
            }

            if (result.typeCode) {
                document.typeCode = parseInt(result.typeCode);
            }

            if(testReport.urlId) {
                document.urlId = testReport.urlId;
            }

            document.sectionId = params.sectionId;
            document.created = new Date();

            var date = new Date();
            document.updated = new Date();
            document.updatedMillis = date.getTime();

            documents.push({index: {_index: instance.config.es.index, _type: instance.config.es.indexType}})
            documents.push(document);
        }

        var result = _(documents)
            .groupBy('testId')
            .map(function(item, itemId) {
                var obj = {};
                obj[itemId] = _.countBy(item, 'type')
                return obj
            }).value();

        var codeResult = _(documents)
            .groupBy('testId')
            .map(function(item, itemId) {
                var obj = {};
                console.log("item.........................", item)
                obj[itemId] = _.countBy(item, 'code')
                return obj
            }).value();
        params.documents = documents;
        params.docResult = result;
        params.codeResult = codeResult;


        var principleResult = _(documents)
            .groupBy('testId')
            .map(function(item, itemId) {
                var obj = {};
                obj[itemId] = _.countBy(item, 'principle')
                return obj
            }).value();

        var psResult = _(documents)
            .groupBy('principle')
            .map(function(item, itemId) {
                var obj = {};
                obj[itemId] = _.countBy(item, 'principleSubTitle')
                return obj
            }).value();

        var levelResult = _(documents)
            .groupBy('level')
            .map(function(item, itemId) {
                var obj = {};
                obj[itemId] = _.countBy(item, 'guidelineIndex')
                return obj
            }).value();

        var pulled = _.pullAt(psResult, [0])
        params.principleSubResults = psResult;

        var pulled1 = _.pullAt(levelResult, [0])
        params.levelResult = levelResult;

        var techniqueResult = _(documents)
            .groupBy('testId')
            .map(function(item, itemId) {
                var obj = {};
                obj[itemId] = _.countBy(item, 'techniqueName')
                return obj
            }).value();

        var totalLevelResults = {};

        if(levelResult.length > 0) {
            _.forEach(levelResult, function(value) {
                var count = 0;
                _.forEach(value, function(val, key) {
                    _.forEach(val, function(val1, key1) {
                        count = count + val1
                    });
                    totalLevelResults[key] = count;
                });
            });
        }

        params.totalLevelResults = totalLevelResults;
        params.principleResult = principleResult;
        params.techniqueResult = techniqueResult;
        callback(null, params);
    }
}

function getAllySummaryResult(params, callback) {
    var searchParam = _getQuery(params);
    console.info('searchParam = ', searchParam);

    params.es.search(searchParam, function(err, response) {
        params.summaryDocument = response;
        console.log("response ", response)
        callback(err, params);
    });
}

function _getQuery(params) {
    var range = {};
    var filterData = {};
    if(params.updated) {
        range.gt = params.updated;
        range.time_zone = '+05:30'
    }

    var options = params.options;
    var instance = params.instance;
    var constants = params.constants;
    var testReport = params.testReport;

    var body = {
        "size": 1,
        "sort":[ { "updated":{ "order":"asc"  }  } ],
        "query": {
            "bool": {
                "must": [
                    { "match": { "testExecutionId":   instance.config[constants.CONFIG_TEST_EXECUTION_ID]        }},
                    { "match": { "stats.status": "COMPLETED" }},
                    { "range":{ "updated": range } }
                ]
            }
        },
        "aggs":{
            "agg_sum_warning": {  "sum":{ "field":"stats.warning" }  },
            "agg_sum_notice":{ "sum":{ "field":"stats.notice" } },
            "agg_sum_error":{ "sum":{ "field":"stats.error" } }
        }
    };

    return {
        index: options.index,
        type: options.summaryIndexType,
        body: JSON.stringify(body)
    };

}

function _getSummaryDocument(params, callback) {

    var constants = params.constants;
    var instance = params.instance;
    var testReport = params.testReport;
    var results = testReport.result;
    var documents = params.documents;
    document = params.summaryDocument.hits.hits[0]._source;
    var standard = document.stats;
    standard.total = results.length;
    standard.result = params.docResult[1][instance.config.testId];
    standard.codeResult = params.codeResult[1][instance.config.testId];
    standard.principleResult = params.principleResult[1][instance.config.testId];
    standard.techniqueResult = params.techniqueResult[1][instance.config.testId];
    standard.principleSubResults = params.principleSubResults;
    standard.levelResult = params.levelResult;
    standard.totalLevelResults = params.totalLevelResults;
    standard.started = testReport.started.getTime();
    if (testReport.status) {
        standard.status = testReport.status;
    }

    var ended = new Date();
    standard.ended = ended.getTime();

    document.stats = standard;
    var date = new Date();
    document.updated = date;
    document.updatedMillis = date.getTime();
    documents.push({update: {_id : params.summaryDocument.hits.hits[0]._id, _index: instance.config.es.index, _type: instance.config.es.summaryIndexType}})
    documents.push({doc: document});


    params.documents = documents;
    callback(null, params);
}

function _createSummaryDocument(params, callback) {

    var constants = params.constants;
    var instance = params.instance;
    var testReport = params.summaryReport;
    //var total = params.summaryDocument.hits.total;

    var documents = [];
    var document = {};
    document.testExecutionId = instance.config[constants.CONFIG_TEST_EXECUTION_ID];
    document.userId = instance.config[constants.CONFIG_TEST_USER_ID];
    document.userName = instance.config[constants.CONFIG_TEST_USER_NAME];
    document.projectId = instance.config.projectId;
    document.componentId = instance.config.applicationId;
    document.projectName = instance.config.projectName;
    document.projectDescription = instance.config.projectDescription;
    document.testId = instance.config.testId;
    document.testName = instance.config.testName;
    document.environmentId = instance.config.environmentId;
    document.environmentName = instance.config.environmentName;
    document.testExecutionName = instance.config.testExecutionName;
    document.measurementName = instance.config.measurementName;

    if (testReport.script) {
        document.viewport = testReport.script.config.page.viewport;
    }

    var stats = [];
    var standard = {};
    standard.urlId = testReport.urlId;
    standard.scenarioId = testReport.scenarioId;
    standard.sectionId = testReport.sectionId;
    standard.standard = testReport.standard;
    standard.url = testReport.requestUrl.url;
    standard.headers = testReport.requestUrl.headers;
    standard.started = testReport.started.getTime();

    if (testReport.scenario) {
        standard.scenario = testReport.scenario.name;
    }

    if (testReport.flow) {
        standard.method = testReport.flow.method;
    }

    if (testReport.status) {
        standard.status = testReport.status;
    }

    var ended = new Date();
    standard.ended = ended.getTime();

    //stats.push(standard);
    document.stats = standard;

    document.created = new Date();

    var date = new Date();
    document.updated = new Date();
    document.updatedMillis = date.getTime();

    documents.push({index: {_index: instance.config.es.index, _type: instance.config.es.summaryIndexType}})
    documents.push(document);
    params.documents = documents;
    callback(null, params);
}

function _publishDocument(params, callback) {
    var documents = params.documents;
    var doc = _.chunk(documents, 1000)
    var docCount = 0;
    if (documents.length > 1) {
        for(var i = 0; i < doc.length; i++) {
            params.es.bulk({
                body: doc[i]
            }, function (err, resp) {
                params.resp = resp;
                console.info('result = ', resp);
                console.info('err = ', err);
            });
        }
        callback(null, params);

    } else {
        callback(null, params);
    }
}



function _checkIndexExsits(params, callback) {
    params.esReporter.indexExists(function(err, response) {
        params.indexExists = response;
        callback(err, params);
    })
}

function _checkAndCreateMapping(params, callback) {
    if(!params.response) {
        params.esReporter.putMapping(function(err, response) {
            params.mappingExists = response;
            callback(err, params);
        });
    } else {
        params.mappingExists = true;
        callback(null, params);
    }
}

function getMapping(options) {
    var body = {
        mappings: {
            ACCESSIBILITY: {
                properties: {
                    project_id: { type: "string"},
                    component_id: { type: "string" },
                    group_id: { type: "string" },
                    project_name: { type: "string" },
                    project_description: { type: "string" },
                    environment: { type: "string" },
                    test_id: { type: "string" },
                    test_execution_id : { type: "string" },
                    test_type: { type: "string" },
                    execution_time : { type: "integer" },
                    user_id: { type: "string" },
                    user_name: { type: "string" },
                    version : { type: "float" },
                    test_url : { type: "string" },
                    standard: { type: "string" },
                    title: { type: "string" },
                    description: { type: "string" },
                    testing: { type: "string" },
                    level: { type: "string" },
                    principle: { type: "string" },
                    principleSubTitle: { type: "string" },
                    guideline: { type: "string" },
                    guidelineIndex: { type: "string" },
                    techniques: { type: "string" },
                    echniqueName: { type: "string" },
                    object: { type: "string" },
                    code: { type: "string" },
                    context: { type: "string" },
                    message: { type: "string" },
                    selector: { type: "string" },
                    type: { type: "string" },
                    urlId: {type: "string" },
                    sectionId: {type: "string" },
                    typeCode: { type: "integer" },
                    start_time: {type: "date", format: "YYYY-MM-DD'T'HH:mm:ssZ"},
                    end_time: {type: "date", format: "YYYY-MM-DD'T'HH:mm:ssZ"},
                    status: { type: "boolean", index : "not_analyzed" }
                }
            }
        }
    };

    return {
        index: options.index,
        body: JSON.stringify(body)
    };
};

function getSummaryMapping(options) {
    var body = {
        mappings: {
            ACCESSIBILITY_SUMMARY: {
                properties: {
                    project_id: { type: "string"},
                    component_id: { type: "string" },
                    group_id: { type: "string" },
                    project_name: { type: "string" },
                    project_description: { type: "string" },
                    environment: { type: "string" },
                    test_id: { type: "string" },
                    test_execution_id : { type: "string" },
                    test_type: { type: "string" },
                    execution_time : { type: "integer" },
                    user_id: { type: "string" },
                    user_name: { type: "string" },
                    version : { type: "float" },
                    test_url : { type: "string" },
                    standard: { type: "string" },
                    status: { type: "string" },
                    viewport: {
                        type: "nested"
                    },
                    stats: {
                        type: "nested",
                        index : "not_analyzed"
                    },
                    start_time: {type: "date", format: "YYYY-MM-DD'T'HH:mm:ssZ"},
                    end_time: {type: "date", format: "YYYY-MM-DD'T'HH:mm:ssZ"},
                    status: { type: "boolean", index : "not_analyzed" }
                }
            }
        }
    };

    return {
        index: options.index,
        body: JSON.stringify(body)
    };
};


function getSummarySearchParam(params) {
    var options = params.options;
    var instance = params.instance;
    var constants = params.constants;
    var testReport = params.testReport;

    var body = {
            "size": 1,
            "query": {
                "bool": {
                    "must": [
                        { "match": { "testExecutionId":   instance.config[constants.CONFIG_TEST_EXECUTION_ID]        }},
                        { "match": { "stats.sectionId": testReport.sectionId }}
                    ]
                }
            }
        };

    return {
        index: options.index,
        type: options.summaryIndexType,
        body: JSON.stringify(body)
    };
};

function getPrincipleSubTitle(guidelineIndex) {
    var principleSubTitle = "";
    if(guidelineIndex.startsWith("1_1")) {
        principleSubTitle = esConstants.GUIDELINE_1_1;
    } else if(guidelineIndex.startsWith("1_2")) {
        principleSubTitle = esConstants.GUIDELINE_1_2;
    } else if(guidelineIndex.startsWith("1_3")) {
        principleSubTitle = esConstants.GUIDELINE_1_3;
    } else if(guidelineIndex.startsWith("1_4")) {
        principleSubTitle = esConstants.GUIDELINE_1_4;
    } else if(guidelineIndex.startsWith("2_1")) {
        principleSubTitle = esConstants.GUIDELINE_2_1;
    } else if(guidelineIndex.startsWith("2_2")) {
        principleSubTitle = esConstants.GUIDELINE_2_2;
    } else if(guidelineIndex.startsWith("2_3")) {
        principleSubTitle = esConstants.GUIDELINE_2_3;
    } else if(guidelineIndex.startsWith("2_4")) {
        principleSubTitle = esConstants.GUIDELINE_2_4;
    } else if(guidelineIndex.startsWith("3_1")) {
        principleSubTitle = esConstants.GUIDELINE_3_1;
    } else if(guidelineIndex.startsWith("3_2")) {
        principleSubTitle = esConstants.GUIDELINE_3_2;
    } else if(guidelineIndex.startsWith("3_3")) {
        principleSubTitle = esConstants.GUIDELINE_3_3;
    } else if(guidelineIndex.startsWith("4_1")) {
        principleSubTitle = esConstants.GUIDELINE_4_1;
    }

    return principleSubTitle;
}

function getTechnique(technique) {
    if(technique.startsWith("G")) {
        technique = esConstants.GENERAL;
    } else if (technique.startsWith("H")) {
        technique = esConstants.HTML;
    } else if (technique.startsWith("B")) {
        technique = technique;
    } else if (technique.startsWith("C")) {
        technique = esConstants.CSS;
    } else if (technique.startsWith("SCR")) {
        technique = esConstants.SCRIPT;
    } else if (technique.startsWith("SVR")) {
        technique = esConstants.SERVER;
    } else if (technique.startsWith("SM")) {
        technique = esConstants.SMIL;
    } else if (technique.startsWith("T")) {
        technique = esConstants.TEXT;
    } else if (technique.startsWith("ARIA")) {
        technique = esConstants.ARIA;
    } else if (technique.startsWith("F")) {
        technique = esConstants.COMMON_FAILURES;
    }

    return technique;
}

function isIndexExists(es, params, callback) {
    es.indices.exists(params, function(error, response){
        callback(error, response);
    });
}

function ping(es, params, callback) {
    es.ping(params, function (error, response, status) {
        callback(error, response, status);
    });
}

module.exports = ElasticClient;