/**
 * Created by senthil on 28/07/17.
 */
'use strict';

var uuid = require('uuid'),
    util = require('util'),
    _ = require('lodash'),
    ElasticClient = require('./es/es.client'),
    a11yConstants = {
        PLUGIN_NAME: 'elasticsearch',

        // indexes of artillery's results:
        TIMESTAMP: 0,
        REQUEST_ID: 1,
        LATENCY: 2,
        STATUS_CODE: 3,

        // required configuration names
        CONFIG_TEST_NAME: 'testName',
        CONFIG_MEASUREMENT_NAME: 'measurementName',
        CONFIG_ERROR_MEASUREMENT_NAME: 'errorMeasurementName',
        CONFIG_TEST_RUN_ID: 'testRunId',
        CONFIG_EXCLUDE_TEST_RUN_ID: 'excludeTestRunId',
        CONFIG_STATIC_TAGS: 'tags',
        CONFIG_TEST_EXECUTION_ID : 'testExecutionId',
        CONFIG_TEST_USER_ID : 'userId',
        CONFIG_TEST_USER_NAME : 'userName',
        CONFIG_TEST_ENV_ID : 'environmentId',
        CONFIG_TEST_ENV_NAME : 'environmentName',

        // Defaults
        DEFAULT_MEASUREMENT_NAME: 'latency',
        DEFAULT_ERROR_MEASUREMENT_NAME: 'clientErrors',

        // environment variables
        CONFIG_ES: 'es',
        CONFIG_ES_PROTOCOL: 'protocol',
        CONFIG_ES_HOST: 'host',
        CONFIG_ES_PORT: 'port',
        CONFIG_ES_USERNAME: 'username',
        CONFIG_ES_PASSWORD: 'password',
        CONFIG_ES_INDEX: 'index',
        CONFIG_ES_INDEX_TYPE: 'indexType',
        CONFIG_ES_INDEX_AVAILABLE: 'indexAvailable',
    },
    a11yMessages = {
        pluginsConfigNotFound: 'No "plugins" configuration found.',
        pluginConfigIsRequired: 'The configuration for %s is required.',
        pluginParamIsRequired: 'The configuration parameter %s is required.',
        pluginParamOrEnvIsRequired: 'The configuration parameter %s or environment variable %s is required.',
        esHostMustBeHostname: 'The %s property must be a host name only, protocol and port cannot be used.',
        metricsReportedToES: '%s metrics reported to ElasticSearch.'
    },
    a11yImpl = {
        handleError: function(message) {
            throw new Error(message);
        },
        validateConfig: function (scriptConfig) {
            // These are the minimum required config values

            var requiredESConfigs = [
                a11yConstants.CONFIG_ES_HOST,
                a11yConstants.CONFIG_ES_INDEX
            ];

            // There must be a configuration object.
            if (!scriptConfig) {
                a11yImpl.handleError(util.format(a11yMessages.pluginConfigIsRequired, a11yConstants.PLUGIN_NAME));
            }

            // Create a set of static tags if none already
            if (!scriptConfig[a11yConstants.CONFIG_STATIC_TAGS]) {
                scriptConfig[a11yConstants.CONFIG_STATIC_TAGS] = {};
            }

            // It must provide a test name; alternate version accepted.
            if (!scriptConfig[a11yConstants.CONFIG_TEST_NAME]) {
                a11yImpl.handleError(util.format(a11yMessages.pluginParamIsRequired, a11yConstants.CONFIG_TEST_NAME));
            }

            // Add the test name to the set of tags to be written
            scriptConfig[a11yConstants.CONFIG_STATIC_TAGS][a11yConstants.CONFIG_TEST_NAME] = scriptConfig[a11yConstants.CONFIG_TEST_NAME];

            // Use measurement names provided in the config, otherwise use the default ones.
            if (!scriptConfig[a11yConstants.CONFIG_MEASUREMENT_NAME]) {
                scriptConfig[a11yConstants.CONFIG_MEASUREMENT_NAME] = a11yConstants.DEFAULT_MEASUREMENT_NAME;
            }

            if (!scriptConfig[a11yConstants.CONFIG_ERROR_MEASUREMENT_NAME]) {
                scriptConfig[a11yConstants.CONFIG_ERROR_MEASUREMENT_NAME] = a11yConstants.DEFAULT_ERROR_MEASUREMENT_NAME;
            }

            // If no testRunId is provided in the static tags, and excludeTestRunId is not set then generate one.
            if (!scriptConfig[a11yConstants.CONFIG_STATIC_TAGS][a11yConstants.CONFIG_TEST_RUN_ID] &&
                !scriptConfig[a11yConstants.CONFIG_EXCLUDE_TEST_RUN_ID]) {
                scriptConfig[a11yConstants.CONFIG_STATIC_TAGS][a11yConstants.CONFIG_TEST_RUN_ID] = uuid.v4();
            }

            // Check each of the elasticsearch-specific settings and validate.
            requiredESConfigs.forEach(function(configName) {
                    if (!scriptConfig[a11yConstants.CONFIG_ES][configName]) {
                        a11yImpl.handleError(util.format(a11yMessages.pluginParamIsRequired, a11yConstants.CONFIG_ES + '.' + configName));
                    }

                    // Check that host is name only: no protocol or port.
                    if (configName === a11yConstants.CONFIG_ES_HOST) {
                        var host = scriptConfig[a11yConstants.CONFIG_ES][configName];
                        if (host.indexOf(':') > -1 || host.indexOf('/') > -1) {
                            a11yImpl.handleError(util.format(a11yMessages.esHostMustBeHostname, a11yConstants.CONFIG_ES + '.' + configName));
                        }
                    }
                }
            );

            return scriptConfig;
        },
        createElasticSearchReporter: function (config) {
            return new ElasticClient(config.es);
        },
        esReportResults: function (instance, esReporter, testReport) {

            var params = {};
            params.instance = instance;
            params.testReport = testReport;
            params.constants = a11yConstants;
            //console.info('*********** testReport = ', testReport);
            esReporter.publishDocument(params, function (err, params) {
                if (err) {
                    a11yImpl.handleError(err.message);
                } else {
                    //console.log(util.format(a11yMessages.metricsReportedToES, points.length));
                }
            });
        },

        getSummaryResult: function (instance, esReporter, testReport, io1) {

            var params = {};
            params.instance = instance;
            params.testReport = testReport;
            params.constants = a11yConstants;
            //console.info('*********** testReport = ', testReport);
            esReporter.getA11ySummaryResults(params, function (err, params) {
                if (err) {
                    a11yImpl.handleError(err.message);
                } else {
                    var privateRomm = '#usernameRoom';
                    //console.log(util.format(a11yMessages.metricsReportedToES, points.length));
                }
            });
        },

        esSummaryStats: function (instance, esReporter, report) {

            var params = {};
            params.instance = instance;
            params.summaryReport = report;
            params.constants = a11yConstants;
            esReporter.createSummaryDocument(params, function (err, params) {
                if (err) {
                    a11yImpl.handleError(err.message);
                } else {
                    //console.log(util.format(a11yMessages.metricsReportedToES, points.length));
                }
            });
        },
        reportErrors: function (instance, esReporter, testReport) {
            var errorCount = 0,
                points = [];

            // Work around change in testReport schema in artillery-core (Issue #3)
            if (testReport._errors) {
                testReport.errors = testReport._errors;
            }

            // If there are no errors or error measurement name not defined (or empty), then exit.
            if (!testReport.errors || !instance.config[a11yConstants.CONFIG_STATIC_TAGS]) {
                return;
            }

            // For each of the error types reported, create a point for influx.
            Object.getOwnPropertyNames(testReport.errors).forEach(function(propertyName) {
                errorCount += testReport.errors[propertyName];
            });

            if (errorCount > 0) {
                // If errors were found, log them.
                var params = {};
                params.instance = instance;
                params.testReport = testReport;
                params.a11yConstants = a11yConstants;
                params.errorCount = errorCount;

                esReporter.publishDocument(params, function (err, params) {
                    if (err) {
                        a11yImpl.handleError(err.message);
                    } else {
                        //console.log(util.format(a11yMessages.metricsReportedToES, points.length));
                    }
                });
            }
        }
    },
    a11yApi = {
        init: function (scriptConfig, eventEmitter) {
            var reporter,
                esReporter,
                that;

            that = this;
            //console.info('scriptConfig = ', scriptConfig);
            // Check for presence of any Plugin configuration.
            if (!scriptConfig || !scriptConfig.plugins) {
                a11yImpl.handleError(a11yConstants.pluginsConfigNotFound);
            }

            // Validate the settings provided for our specific Plugin.
            this.config = a11yImpl.validateConfig(scriptConfig.plugins[a11yConstants.PLUGIN_NAME]);

            // Create a reporting client and attach to DONE event from Artillery.
            esReporter = a11yImpl.createElasticSearchReporter(this.config);

            esReporter.indexExists(function(err, response) {
                if(!response) {
                    esReporter.putMapping(function(err, response) {
                        that.config[a11yConstants.CONFIG_ES_INDEX_AVAILABLE] = response;
                    })
                } else {
                    that.config[a11yConstants.CONFIG_ES_INDEX_AVAILABLE] = true;
                }
            });

            eventEmitter.on('a11ystats', function (report) {
                // Got some results! Report results to ElasticSearch.
                a11yImpl.esReportResults(that, esReporter, report);
                //a11yImpl.reportErrors(that, esReporter, report);
            });

            eventEmitter.on('a11ysummarystats', function (report) {
                // Got some results! Report results to ElasticSearch.
                a11yImpl.esSummaryStats(that, esReporter, report);
                //a11yImpl.reportErrors(that, esReporter, report);
            });
        }
    };

module.exports = a11yApi.init;

/* test-code */
module.exports.a11yConstants = a11yConstants;
module.exports.a11yMessages = a11yMessages;
module.exports.a11yImpl = a11yImpl;
module.exports.a11yApi = a11yApi;
/* end-test-code */
